
/* MD5DEEP - helpers.c
 *
 * By Jesse Kornblum
 *
 * This is a work of the US Government. In accordance with 17 USC 105,
 * copyright protection is not available for any work of the US Government.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "md5deep.h"

/* On non-glibc systems we need to compute our name. While this can
   ostensibly be found in argv[0], we clean up this value to make
   it look nicer. The argument s should be argv[0] passed from main(). */
#ifndef __GLIBC__
void setProgramName(char *s) {

  /* 's' should be argv[0]. Our goal is get the real filename for this
     value and then strip out everything before the last DIR_TRAIL_CHAR */
  int pos;
  char *fn = (char *)malloc(sizeof(char) * PATH_MAX);
  if (realpath(s,fn) == NULL) {
    __progname = strdup("md5deep");
    return;
  }
  
  pos = strlen(fn);

  while (fn[pos] != DIR_TRAIL_CHAR && pos > 0)
    pos--;

  /* If we don't find a DIR_TRAIL_CHAR, the whole real path is the 
     name of the program */
  if (pos == 0) {
    if (fn[0] == DIR_TRAIL_CHAR)
      fn++;
  } else {

    /* Advance one character past the DIR_TRAIL_CHAR that we found */
    fn += (++pos);
  }

  __progname = fn;
}

#endif /* ifndef __GLIBC__ */






#if defined (__UNIX)

/* Return the size, in bytes of an open file stream. On error, return -1 */
#if defined (__LINUX)

unsigned long long measureOpenFile(FILE *f) {

  int descriptor = 0;
  struct stat *info;

  /* I don't know why, but if you don't initialize this value you'll
     get wildly innacurate results when you try to run this function */
  unsigned long long numsectors = 0;
  unsigned long long total = 0, original = ftello(f);

  if ((fseeko(f,0,SEEK_END)))
    return -1;
  total = ftello(f);
  if ((fseeko(f,original,SEEK_SET)))
    return -1;

  /* Block devices, like /dev/hda, don't return a normal filesize.
     If we are working with a block device, we have to ask the operating
     system to tell us the true size of the device. 
     
     The following only works on Linux as far as I know. If you know
     how to port this code to another operating system, please contact
     the current maintainer of this program! */

  descriptor = fileno(f);
  info = (struct stat*)malloc(sizeof(struct stat));

  /* I'd prefer not to use fstat as it will follow symbolic links. We don't
     follow symbolic links. That being said, all symbolic links *should*
     have been caught before we got here. */

  fstat(descriptor,info);
  if (S_ISBLK(info->st_mode)) {
    if (ioctl(descriptor, BLKGETSIZE, &numsectors)){
#if defined(__DEBUG)
      perror("BLKGETSIZE failed");
#endif
    } else {

      /* We're going to assume that this device has 512 byte sectors.
	 Eventually we should add a way to get the real number of sectors
	 from the device. */
      total = numsectors * 512;
    }
  }
  
  free(info);
  return (total - original);

}

#elif defined (__MACOSX)

#include <stdint.h>
#include <sys/ioctl.h>
#include <sys/disk.h>

unsigned long long measureOpenFile(FILE *f) {
  struct stat info;
  off_t total = 0;
  off_t original = ftello(f);
  int ok = TRUE, fd = fileno(f);


  /* I'd prefer not to use fstat as it will follow symbolic links. We don't
     follow symbolic links. That being said, all symbolic links *should*
     have been caught before we got here. */

  fstat(fd, &info);

  /* Block devices, like /dev/hda, don't return a normal filesize.
     If we are working with a block device, we have to ask the operating
     system to tell us the true size of the device. 
     
     The following only works on Linux as far as I know. If you know
     how to port this code to another operating system, please contact
     the current maintainer of this program! */

  if (S_ISBLK(info.st_mode)) {
	daddr_t blocksize = 0;
	daddr_t blockcount = 0;


    /* Get the block size */
    if (ioctl(fd, DKIOCGETBLOCKSIZE,blocksize) < 0) {
      ok = FALSE;
#if defined(__DEBUG)
      perror("DKIOCGETBLOCKSIZE failed");
#endif
    } 
  
    /* Get the number of blocks */
    if (ok) {
      if (ioctl(fd, DKIOCGETBLOCKCOUNT, blockcount) < 0) {
#if defined(__DEBUG)
	perror("DKIOCGETBLOCKCOUNT failed");
#endif
      }
    }

    total = blocksize * blockcount;

  }

  else {

    /* I don't know why, but if you don't initialize this value you'll
       get wildly innacurate results when you try to run this function */

    if ((fseeko(f,0,SEEK_END)))
      return -1;
    total = ftello(f);
    if ((fseeko(f,original,SEEK_SET)))
      return -1;
  }

  return (total - original);
}



#else 

/* This is code for general UNIX systems 
   (e.g. NetBSD, FreeBSD, OpenBSD, etc) */

unsigned long long measureOpenFile(FILE *f){

  unsigned long long total = 0, original = ftello(f);

  if ((fseeko(f,0,SEEK_END)))
    return -1;
  total = ftello(f);
  if ((fseeko(f,original,SEEK_SET)))
    return -1;
  
  return (total - original);
}

#endif /* UNIX Flavors */

#endif  /* ifdef __UNIX */





#if defined(__WIN32)
unsigned long long measureOpenFile(FILE *f){

  unsigned long long total = 0, original = ftello(f);

  if ((fseeko(f,0,SEEK_END)))
    return -1;
  total = ftello(f);
  if ((fseeko(f,original,SEEK_SET)))
    return -1;
  
  return (total - original);
}
#endif /* ifdef __WIN32 */

